#!/usr/bin/python

"""

pyDox v0.5a

Seung Chan Lim ( dev@djslim.com )

8/23/2001

:: Lazy attempt to automate python API documentation

This will basically print out a somewhat nicely indented text file to stdout.
So REDIRECT it to some text file if you want to grab it.

It uses python docstrings for functions, classes and modules.

You can also control whether or not to print out the private, protected, and
overloaded operator methods of a class.

The category of things that the script handles are:

 1. imported modules in the global name space
 2. globals in the global name space
        a. if the variable name is all-caps, it thinks it's a constant
 3. classes in the global name space
 	a. the constructor
 	b. its public, private, protected and overloaded operator methods
	c. its destructor
 	d. tells you the first level parent object
 4. functions in the global name space
 5. classes that are subclasses of the Exception class is handled seprately

Usage: 
	python pyDox.py path_to_the_script
	python pyDox.py path_to_the_script > document.txt
	python pyDox.py path_to_the_script | more
	python pyDox.py path_to_the_the_script path_to_the_XSL_file
"""

__version__  = "0.5a"
__email__    = "dev@djslim.com"
__author__   = "Seung Chan Lim"
__homepage__ = "http://www.djslim.com/"


###############################################################################
#				modules
###############################################################################
import imp
import os.path
import string
import inspect
import re

###############################################################################
#				globals
###############################################################################

# list of the names of special methods that classes can implement
# to emulate operator overloading
special_methods = ["__repr__", "__str__", "__cmp__", "__rcmp__",
                   "__hash__", "__nonzero__", "__getattr__", "__setattr__",
                   "__delattr__", "__call__", "__len__", "__getitem__",
                   "__setitem__", "__delitem__", "__setslice__", 
                   "__delslice__", "__getslice__", "__contains__", "__add__",
                   "__sub__ ", "__mul__", "__div__", "__mod__", "__divmod__",
                   "__pow__", "__lshift__", "__rshift__", "__and__", "__xor__",
                   "__or__"]

# names of special global variables that pyDox treats specially
special_vars    = ["__version__", "__author__", "__homepage__", "__email__"]

stripIt         = string.strip
joinIt          = string.join
splitIt         = string.split

# regular expression to strip '#' from comment chunks
comment_regex   = re.compile("^\s*#+(.+)")




###############################################################################
# wraps text with the CDATA tag
###############################################################################
def __cdatafy(text):
    
    return "<![CDATA[%s]]>"%(str(text))


###############################################################################
# replaces &, <, > with their ampersand equivalent
###############################################################################
def __xmlEncode(text):
    
    text = string.replace(text, '&', '&amp;')
    text = string.replace(text, '<', '&lt;')
    text = string.replace(text, '>', '&gt;')
    
    return text


###############################################################################
# returns am XML representation of a native python variable
###############################################################################
def __xmlifyVar(var_name, var):
    
    xml_type_tag = string.replace(str(type(var)), "<type '", "")
    xml_type_tag = string.lower(string.replace(xml_type_tag, "'>", ""))
    
    return stripIt("""
<name>%s</name>
<val type="%s">%s</val>
"""%(var_name,
     xml_type_tag,
     __xmlEncode(repr(var))))
     
     
###############################################################################
# formats lines of string so that each line is stripped off of the amount
# of white space the first line is indented by
###############################################################################
def __formatDocString(docstr):
    
    if not docstr:
        return ""
    
    # using the first line as the tab point we delete
    # at most that much white space from the following lines
    docstr = string.expandtabs(docstr)
    
    indent = len(docstr) - len(string.lstrip(docstr))
    docstr_list     = splitIt(docstr,"\n")
    docstr_list_len = len(docstr_list)
    
    for i in range(docstr_list_len):
        total_stripped = len(docstr_list[i]) - \
                                 len(string.lstrip(docstr_list[i]))
                                 
        docstr_list[i] = docstr_list[i][min(total_stripped, indent):]
            
    # get rid of 0 length strings 
    # docstr_list = filter(lambda i: len(i) > 0, docstr_list)
    
    return stripIt("%s\n"%(joinIt(docstr_list,"\n")))
    

###############################################################################
# formats a given comment string and strips the preceeding "#"s and line
# that only contain "#"s
###############################################################################
def __formatCommentString(commentstr):
    
	tmp_str = splitIt(commentstr, "\n")
	
	# for each line I want the "#" removed until a different character 
	# has been found
	tmp_str_len = len(tmp_str)
	
	for i in range(tmp_str_len):
		regex_match = comment_regex.match(tmp_str[i])
		if regex_match:
			tmp_str[i] = regex_match.group(1)
	
	tmp_str = filter(lambda i: i != "#", tmp_str)
	
	return joinIt(tmp_str,"\n")
		
	
###############################################################################
# returns an XML representation of the arguments of a function
###############################################################################
def __xmlifyFuncArgs(func):

    normal_arg_names  = []
    default_arg_names = []
    special_arg_names = []
    
    try:
        func_code = func.func_code
    except AttributeError:
        # couldn't retrieve the python code of the function
        normal_arg_names = ["<argument type=\"normal\">\n <name>...</name>\n</argument>"]
    else:
        arg_names, args, kw, defaults = inspect.getargspec(func)
    
        # take care of the arguments w/ no default val
        arg_count = len(arg_names)
	
	try:
	    def_arg_count = len(defaults)
	except TypeError:
	    def_arg_count = 0
	    
        first_def_arg_index = arg_count - def_arg_count

	# the normal arguments
        for i in range(arg_count - def_arg_count):
            arg_names[i] = "<argument>\n <name>%s</name>\n</argument>\n"%(arg_names[i])

        # take care of the arguments w/ default val
        for i in range(def_arg_count):
            arg_names[first_def_arg_index + i] ="<argument>\n %s\n</argument>\n"%(\
                    __xmlifyVar(arg_names[first_def_arg_index + i], defaults[i]))

        normal_arg_names  = arg_names[:first_def_arg_index]
        default_arg_names = arg_names[first_def_arg_index:]
        
	# if it uses *args, we should include that, too
        if args:
            special_arg_names.append("<argument type=\"list\">\n <name>%s</name>\n</argument>\n"%(args))

	# if it uses **kw, we should include that, too
        if kw:
            special_arg_names.append("<argument type=\"keyword\">\n <name>%s</name>\n</argument>\n"%(kw))
        
    return stripIt("""
<arguments defaultIndex="%d" argumentCount="%d">
%s
%s
%s
</arguments>
"""%(first_def_arg_index + 1,
     arg_count,
     stripIt(joinIt(normal_arg_names,"")),
     stripIt(joinIt(default_arg_names,"")),
     stripIt(joinIt(special_arg_names,""))))


###############################################################################
# filters the namespace passed in by categories such as functions,
# variables, classes, and modules
###############################################################################
def __filterNameSpace(namespace):
    
    assert( type(namespace) == type({}) )
    
    class_dict     = {}
    function_dict  = {}
    module_dict    = {}
    global_dict    = {}
    exception_dict = {}
    
    keys = namespace.keys()
    
    for key in keys:
        key_type = type(namespace[key])
        
        if inspect.isclass(namespace[key]):
            # if it's a class but it's an exception
            if issubclass(namespace[key],Exception):
                exception_dict[key] = namespace[key]
            else:
                class_dict[key]   = namespace[key]
        elif inspect.isfunction(namespace[key]):
            # if it's a function
            function_dict[key] = namespace[key]
        elif inspect.ismodule(namespace[key]):
            module_dict[key]   = namespace[key]
        elif not inspect.isbuiltin(namespace[key]):
            global_dict[key]   = namespace[key]

    return (global_dict,  
            module_dict, 
            function_dict, 
            class_dict, 
            exception_dict)
    

###############################################################################
# returns an XML representation of the baes class(es) the class passed in
# inherits from
###############################################################################
def __xmlifyBaseClasses(class_obj):
    
    class_obj_parents = class_obj.__bases__
    
    # this class is the base class
    if len(class_obj_parents) == 0:
    
        return ""

    hierarchy_text = []
    
    for parent in class_obj_parents:
    
        # for some reason the module name defaults to dummy
        # if itself is the module... odd
        if parent.__module__ != "dummy":
            module = parent.__module__
        else:
            module = ""    
            
        hierarchy_text.append(stripIt("""
<class>
 <name>%s</name>
 <package>%s</package>
</class>
"""%(parent.__name__,
      module)))
        
    return joinIt(hierarchy_text,"\n")
    
    
###############################################################################
# retrieves the doc string if available, if not returns the comment
# that preceeds the object or "n/a" in case neither is available
###############################################################################
def __retrieveDocOrComment(obj):
	if hasattr(obj, "__doc__"):
		if obj.__doc__ != None:
		
			return __formatDocString(obj.__doc__)
	
	comment = inspect.getcomments(obj) or ""
        return __formatCommentString(comment) or "n/a"
		
		
###############################################################################
# 
###############################################################################
def loadScript(script_path):
    """
    reads script from the path and returns namespace in which the script 
    got executed
    """
    
    try:
        fo = open(script_path,"r")
    except IOError:
        raise IOError, "The script you wanted to document could not be\
opened"
    else:
        script_content = fo.read()
        fo.close()
        
        try:
            script_content = compile(script_content, script_path, \
                                     "exec")
            
            tmp_module = imp.new_module("dummy")
            tmp_module.__file__ = os.path.basename(script_path)
            
            exec script_content in tmp_module.__dict__

        except:
            import sys
            import traceback
            
            sys.stderr.write("""
*******************************************************************************

There was an error interpreting the script you want documented.
Please consult the following traceback!!!

*******************************************************************************

""")
	    traceback.print_exc(sys.stderr)

    
            
            return None
    
            
    return tmp_module
    
    
###############################################################################
# 
###############################################################################
def getScriptInfo(script_obj):
    """
    returns an XML representation of the script info
    """
    tmp_dict             = {}
    tmp_dict["author"]   = ""
    tmp_dict["email"]    = ""
    tmp_dict["homepage"] = ""
    tmp_dict["version"]  = ""
    
    try: tmp_dict["author"]   = __cdatafy(script_obj.__author__)
    except: pass

    try: tmp_dict["email"]    = __cdatafy(script_obj.__email__)
    except: pass

    try: tmp_dict["homepage"] = __cdatafy(script_obj.__homepage__)
    except: pass

    try: tmp_dict["version"]  = script_obj.__version__
    except: pass

    # xml encode all text values
    map(lambda key, x = __xmlEncode, info = tmp_dict: x(str(info[key])),
        tmp_dict.keys())

    if len(tmp_dict["author"]) > 0 or len(tmp_dict["email"]) or len(tmp_dict["homepage"]):
        author_text = """
 <name>%s</name>
 <email>%s</email>
 <homepage>%s</homepage>
"""%(tmp_dict["author"],
     tmp_dict["email"],
     tmp_dict["homepage"])
    else:
        author_text = ""
        
    return stripIt("""
<comment>%s</comment>
<name>%s</name>
<path>%s</path>
<version>%s</version>
<author>
%s
</author>
""" %( __cdatafy(__retrieveDocOrComment(script_obj) or "n/a"),
       inspect.getmodulename(script_obj.__file__),
       __xmlEncode(os.path.abspath(script_obj.__file__)),
       tmp_dict["version"],
       stripIt(author_text)) )


###############################################################################
#
###############################################################################
def getGlobalsInfo(globals_dict):
    """
    returns an XML representation of the globals
    """
    
    if len(globals_dict) == 0:
        return ""
    
    global_var_text   = []
    global_const_text = []
    global_unimp_text = []
    
    global_keys = globals_dict.keys()
    global_keys.sort()


    
    global_keys = filter(lambda i: i[0] != "_", global_keys)
    
    for key in global_keys:
        if key in special_vars:
	    continue
    
        if string.upper(key) == key:
            global_const_text.append("<constants>%s</constants>"%(\
                __xmlifyVar(key, globals_dict[key])))
        else:
            global_var_text.append("<variable>%s</variable>"%(\
                __xmlifyVar(key,globals_dict[key])))
    
    return stripIt("""
%s
%s
"""%(joinIt(global_const_text,"\n"),
     joinIt(global_var_text,"\n")))
    

###############################################################################
#
###############################################################################
def getFuncInfo(func_dict):
    """
    returns an XML representation of the top-level functions
    """
    
    if len(func_dict) == 0:
        return ""
        
    func_text = []
    
    func_keys = func_dict.keys()
    func_keys.sort()
    
    for key in func_keys:
        func         = func_dict[key]
        func_name    = func.__name__
        func_comment = __retrieveDocOrComment(func) or "n/a"
            
        # get a nicely formatted argument list
        func_args    = __xmlifyFuncArgs(func)
        
        func_text.append(stripIt("""
<function>
 <comment>%s</comment>
 <name>%s</name>
 %s
</function>
"""%(__cdatafy(func_comment),
     func_name,
     func_args,)))
     
    return joinIt(func_text,"\n")


###############################################################################
#
###############################################################################
def getModulesInfo(module_dict):
    """
    returns an XML representation of the imported modules
    """
    
    if len(module_dict) == 0:
        return ""
        
    modules_text = []
    
    module_keys = module_dict.keys()
    module_keys.sort()
    
    for module in module_keys:
        try:
            module_path = os.path.basename(module_dict[module].__file__)
        except AttributeError:
            module_path = "built-in module"
            
        modules_text.append(string.lstrip("""
<module>
 <name>%s</name>
 <path>%s</path>
</module>
"""%(module_modules[module].__name__, 
     __xmlEncode(module_path)) )) 
    
    return stripIt(joinIt(modules_text))


###############################################################################
# returns an XML representation of the exception classes
###############################################################################    
def getExceptionInfo(exception_dict):
    
    exception_text = ""
    
    if len(exception_dict) == 0:
    
        return exception_text
        
    execption_text = getClassInfo(exception_dict)
        
    return execption_text
    
    
###############################################################################
#
###############################################################################    
def getClassInfo(class_dict):    
    """
    returns an XML representation of all classes
    
        private methods - starts with "__"
        protected methods - starts with "_"
        overloaded operator methods - stuff like "__add__"
    """
    
    if len(class_dict) == 0:
        return ""

    classes_text = []
    
    class_keys = class_dict.keys()
    class_keys.sort()

        
    for class_key in class_keys:
        class_globals, class_modules, class_funcs, class_classes, \
        class_exceptions = \
                      __filterNameSpace(class_dict[class_key].__dict__)
        
        del class_globals["__doc__"]
        del class_globals["__module__"]
        # now divide the methods into private, public, protected 
        # methods and overloaded operators 
        priv_methods = {}
        pub_methods  = {}
        ovr_methods  = {}
        prot_methods = {}
        destructor   = {}
        constructor  = {}
        
        class_func_keys = class_funcs.keys()
        for key in class_func_keys:
            if class_funcs[key].__name__ != key:
                priv_methods[class_funcs[key].__name__] = \
                                               class_funcs[key]
            elif class_funcs[key].__name__ == "__init__":
                constructor[key] = class_funcs[key]
            elif class_funcs[key].__name__ in special_methods:
                ovr_methods[key]  = class_funcs[key]
            elif class_funcs[key].__name__ == "__del__":
                destructor[key]   = class_funcs[key]
            elif class_funcs[key].__name__[0] == "_":
                prot_methods[key] = class_funcs[key]
            else:
                pub_methods[key]  = class_funcs[key]
                
        class_doc = __retrieveDocOrComment(class_dict[class_key])
	
        classes_text.append(stripIt("""
<class>
 <comment>%s</comment>
 <name>%s</name>
 <bases>%s</bases>
 <locals>%s</locals>
 <constructor>%s</constructor>
 <destructor>%s</destructor>
 <operators>%s</operators>
 <methods><public>%s</public></methods>
 <methods><private>%s</private></methods>
 <methods><protected>%s</protected></methods>
</class>
"""%(__cdatafy(class_doc),
     class_dict[class_key].__name__,
     __xmlifyBaseClasses(class_dict[class_key]),
     getGlobalsInfo(class_globals),
     getFuncInfo(constructor),
     getFuncInfo(destructor),
     getFuncInfo(ovr_methods),
     getFuncInfo(pub_methods),
     getFuncInfo(priv_methods),
     getFuncInfo(prot_methods))))
			    
  
    return "%s"%(joinIt(classes_text,""))
    
    
###############################################################################
# prints the usage
###############################################################################
def printUsage(script_path):
    
    print """
Usage: %s <<target python script>> [ <<target XSL file>> ]
""" %(os.path.basename(script_path))




###############################################################################
#                da main
###############################################################################
if __name__ == "__main__":
    import sys
    
    argv = sys.argv
    argc = len(argv)
    
    if argc > 1:

        if argc == 3:
            xslpath = sys.argv[2]
        else:
            # default xsl - included
            xslpath = "pydox2xhtml.xsl"
        
        module_globals    = {} # the module's global variables
        module_functions  = {} # the module's top-level functions
        module_classes    = {} # the module's classes
        module_modules    = {} # modules imported by the module
        module_exceptions = {} # exceptions included in the module
        
        script_obj = loadScript(argv[1])
        

        if script_obj:
            
            # don't want to actually play with the real name space, so
            # make a copy - probably being overly cautious anyway, though.
            tmp_dict = script_obj.__dict__.copy()
            
            # delete the default stuff in the copy of the namespace
            del tmp_dict["__builtins__"]
            del tmp_dict["__doc__"]
            del tmp_dict["__file__"]
            del tmp_dict["__name__"]
            
            module_globals, module_modules, module_functions,\
            module_classes, module_exceptions = __filterNameSpace(tmp_dict)

    
            print "<?xml version=\"1.0\" ?>"
            print "<?xml:stylesheet type=\"text/xsl\" href=\"%s\"?>"%(xslpath)
            print "<file>"
            print getScriptInfo(script_obj)
            print "<imports>"
            print getModulesInfo(module_modules)
            print "</imports>"
            print "<globals>"
            print getGlobalsInfo(module_globals)
            print "</globals>"
            print "<exceptions>"
            print getExceptionInfo(module_exceptions)
            print "</exceptions>"
            print "<functions>"
            print getFuncInfo(module_functions)
            print "</functions>"
            print "<classes>"
            print getClassInfo(module_classes)
            print "</classes>"
            print "</file>"
            
    elif len(argv) == 1:
    	printUsage(argv[0])
